'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _os = require('os');

var _os2 = _interopRequireDefault(_os);

var _crypto = require('crypto');

var _crypto2 = _interopRequireDefault(_crypto);

var _cluster = require('cluster');

var _cluster2 = _interopRequireDefault(_cluster);

var _http = require('http');

var _http2 = _interopRequireDefault(_http);

var _parameter = require('./parameter');

var _parameter2 = _interopRequireDefault(_parameter);

var _json = require('./json');

var _json2 = _interopRequireDefault(_json);

var _config = require('./config');

var _config2 = _interopRequireDefault(_config);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var signals = ['SIGINT', 'SIGTERM'];
var numCPUs = _os2.default.cpus().length;

function register(store) {
    function registerCommand(data, feed) {
        if (this.isLast()) {
            return feed.close();
        }
        var shasum = _crypto2.default.createHash('sha1');
        shasum.update(data);
        var cmdid = shasum.digest('hex');
        return store.set(cmdid, data).then(function () {
            return feed.send(JSON.stringify(cmdid));
        });
    }
    return registerCommand;
}

function createServer(ezs, store) {
    var startedAt = Date.now();
    var server = _http2.default.createServer(function (request, response) {
        var url = request.url,
            method = request.method,
            headers = request.headers;

        var cmdid = url.slice(1);
        if (url === '/' && method === 'POST') {
            if (headers['x-parameter']) {
                var parameters = _parameter2.default.unpack(headers['x-parameter']);
                _parameter2.default.put(ezs, parameters);
            }
            request.pipe(ezs('concat')).pipe(ezs(register(store))).pipe(response);
        } else if (url.match(/^\/[a-f0-9]{40}$/i) && method === 'POST') {
            store.get(cmdid).then(function (cmds) {
                var processor = void 0;
                try {
                    var commands = _json2.default.parse(cmds);
                    processor = ezs.pipeline(commands);
                } catch (e) {
                    console.log('server cannot execute ' + cmdid, e);
                    response.writeHead(400);
                    response.end();
                    return;
                }
                console.log('server will execute ' + cmdid);
                response.writeHead(200);
                request.pipe(ezs('decoder')).pipe(processor).pipe(ezs.catch(console.error)).pipe(ezs('encoder')).pipe(response);
                request.resume();
            });
        } else if (url === '/' && method === 'GET') {
            store.all().then(function (keys) {
                var info = {
                    concurrency: numCPUs,
                    register: keys.length,
                    uptime: Date.now() - startedAt
                };
                var responseBody = JSON.stringify(info);
                var responseHeaders = {
                    'Content-Type': 'application/json',
                    'Content-Length': Buffer.byteLength(responseBody)
                };
                response.writeHead(200, responseHeaders);
                response.write(responseBody);
                response.end();
            });
        } else {
            response.writeHead(404);
            response.end();
        }
    }).listen(_config2.default.port);
    signals.forEach(function (signal) {
        return process.on(signal, function () {
            return server.close(function () {
                return process.exit(0);
            });
        });
    });
    // console.log(`PID ${process.pid} listening on 31976`);
    return server;
}

function createCluster(ezs, store) {
    var term = false;
    if (_cluster2.default.isMaster) {
        for (var i = 0; i < numCPUs; i += 1) {
            _cluster2.default.fork();
        }
        _cluster2.default.on('exit', function () {
            if (!term) {
                _cluster2.default.fork();
            }
        });
        signals.forEach(function (signal) {
            process.on(signal, function () {
                term = true;
                Object.keys(_cluster2.default.workers).forEach(function (id) {
                    return _cluster2.default.workers[id].kill();
                });
            });
        });
    } else {
        createServer(ezs, store);
    }
    return _cluster2.default;
}

exports.default = {
    createServer: createServer,
    createCluster: createCluster
};